/* sample_milter - example mail filter module
**
** Copyright  2004 by Jef Poskanzer <jef@mail.acme.com>.
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions
** are met:
** 1. Redistributions of source code must retain the above copyright
**    notice, this list of conditions and the following disclaimer.
** 2. Redistributions in binary form must reproduce the above copyright
**    notice, this list of conditions and the following disclaimer in the
**    documentation and/or other materials provided with the distribution.
**
** THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
** ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
** IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
** ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
** FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
** DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
** OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
** HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
** LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
** OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
** SUCH DAMAGE.
**
** For commentary on this license please see http://www.acme.com/license.html
*/

#include <unistd.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <sysexits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/errno.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "libmilter/mfapi.h"


/* Structs. */

/* Per-connection data structure. */
struct connection_data {
    int connection_num;
    int messages;
    int recipients;
    int headers;
    int bodyblocks;
    size_t bodybytes;
    int aborts;
    };


/* Forwards. */

static void usage( void );
static sfsistat sample_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr );
static sfsistat sample_helo( SMFICTX* ctx, char* helohost );
static sfsistat sample_envfrom( SMFICTX* ctx, char** fromargs );
static sfsistat sample_envrcpt( SMFICTX* ctx, char** rcptargs );
static sfsistat sample_header( SMFICTX* ctx, char* name, char* value );
static sfsistat sample_eoh( SMFICTX* ctx );
static sfsistat sample_body( SMFICTX* ctx, unsigned char* bytes, size_t len );
static sfsistat sample_eom( SMFICTX* ctx );
static sfsistat sample_abort( SMFICTX* ctx );
static sfsistat sample_close( SMFICTX* ctx );
static sfsistat sample_unknown( SMFICTX* ctx, const char* cmd );
static sfsistat sample_data( SMFICTX* ctx );
static sfsistat sample_negotiate( SMFICTX* ctx, unsigned long f0, unsigned long f1, unsigned long f2, unsigned long f3, unsigned long* pf0, unsigned long* pf1, unsigned long* pf2, unsigned long* pf3 );


/* Globals. */

static char* argv0;
static int connections;

static struct smfiDesc smfilter =
    {
    "SAMPLE",		/* filter name */
    SMFI_VERSION,	/* version code -- do not change */
    0,			/* flags */
    sample_connect,	/* connection info filter */
    sample_helo,	/* SMTP HELO command filter */
    sample_envfrom,	/* envelope sender filter */
    sample_envrcpt,	/* envelope recipient filter */
    sample_header,	/* header filter */
    sample_eoh,		/* end of header */
    sample_body,	/* body block filter */
    sample_eom,		/* end of message */
    sample_abort,	/* message aborted */
    sample_close,	/* connection cleanup */
    sample_unknown,	/* unrecognized / unimplemented command */
    sample_data,	/* DATA command filter  */
    sample_negotiate	/* negotiation  */
    };


int
main( int argc, char** argv )
    {
    char* sockarg;

    argv0 = strrchr( argv[0], '/' );
    if ( argv0 != (char*) 0 )
	++argv0;
    else
	argv0 = argv[0];

    openlog( argv0, LOG_PERROR, LOG_MAIL );

    connections = 0;

    if ( argc != 2 )
	usage();
    sockarg = argv[1];


    /* Harden our umask so that the new socket gets created securely. */
    umask( 0077 );

    if ( smfi_register( smfilter ) == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_register() failed" );
	exit( EX_OSERR );
	}
    (void) smfi_setconn( sockarg );
    if ( smfi_opensocket( true ) == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_opensocket() failed" );
	exit( EX_OSERR );
	}

    if ( smfi_main() == MI_FAILURE )
	{
	syslog( LOG_ERR, "smfi_main() failed" );
	exit( EX_OSERR );
	}

    syslog( LOG_INFO, "exitting - %d connections", connections );
    closelog();
    exit( EX_OK );
    }


static void
usage( void )
    {
    (void) fprintf( stderr, "usage:  %s socket\n", argv0 );
    exit( EX_USAGE );
    }


/* The filter routines.  These can return the following values:
**
**   SMFIS_CONTINUE	Continue processing the current connection, message,
**			or recipient.
**   SMFIS_REJECT	For a connection-oriented routine, reject this
**			connection; call xxfi_close.
**			For a recipient-oriented routine, reject the current
**			recipient (but continue processing the current message).
**			For a message-oriented routine (except xxfi_eom or
**			xxfi_abort), reject this message.
**   SMFIS_DISCARD	For a message- or recipient-oriented routine, accept
**			this message, but silently discard it.
**			This should not be returned by a connection-oriented
**			routine.
**   SMFIS_ACCEPT	For a connection-oriented routine, accept this
**			connection without further filter processing;
**			call xxfi_close.
**			For a message- or recipient-oriented routine, accept
**			this message without further filtering.
**   SMFIS_TEMPFAIL	Return a temporary failure, i.e. the corresponding
**			SMTP command will return an appropriate 4xx status
**			code.
**			For a connection-oriented routine, fail for this
**			connection; call xxfi_close.
**			For a recipient-oriented routine, only fail for the
**			current recipient; continue message processing.
**			For a message-oriented routine (except xxfi_envfrom),
**			fail for this message.
**
** The milter API is fully documented here:
**   http://www.milter.org/milter_api/
**
** The sequence of callbacks is as follows:
**
**   The SMTP connection is accepted - sample_connect()
**   The client sends a HELO/EHLO command - sample_helo()
**   For each message:
**     The client sends a MAIL FROM command - sample_envfrom()
**     The client sends one or more RCPT TO commands - sample_envrcpt()
**     The client sends a DATA command.
**     The client sends one or more headers - sample_header()
**     The client sends a blank line - sample_eoh()
**     The client sends the message body - zero or more calls to sample_body()
**     The client sends a dot - sample_eom()
**   The client sends a QUIT command - sample_close().
*/


/* sample_connect - handle the initial TCP connection
**
** Called at the start of a connection.  Any per-connection data should
** be initialized here.
**
** connhost: The hostname of the client, based on a reverse lookup.
** connaddr: The client's IP address, based on getpeername().
*/
static sfsistat
sample_connect( SMFICTX* ctx, char* connhost, _SOCK_ADDR* connaddr )
    {
    struct connection_data* cd;
    struct sockaddr_in* sa_in;

    if ( connaddr == (_SOCK_ADDR*) 0 )
	return SMFIS_ACCEPT;	/* can't deal with it */

    cd = (struct connection_data*) malloc( sizeof(struct connection_data) );
    if ( cd == (struct connection_data*) 0 )
	{
	syslog( LOG_ERR, "couldn't allocate connection_data" );
	return SMFIS_ACCEPT;
	}
    (void) smfi_setpriv( ctx, (void*) cd );
    cd->connection_num = ++connections;

    sa_in = (struct sockaddr_in*) ( (void*) connaddr );	/* extra cast to avoid alignment warning */
    syslog( LOG_INFO, "conn# %d - sample_connect( \"%s\" [%s] )", cd->connection_num, connhost, inet_ntoa( sa_in->sin_addr ) );

    cd->messages = 0;
    cd->aborts = 0;

    return SMFIS_CONTINUE;
    }


/* sample_helo - handle the HELO command
**
** Called at the start of a connection.
**
** helohost: The string passed to the HELO/EHLO command.
*/
static sfsistat
sample_helo( SMFICTX* ctx, char* helohost )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_helo( \"%s\" )", cd->connection_num, helohost );

    return SMFIS_CONTINUE;
    }


/* sample_envfrom - handle the MAIL FROM:<> command
**
** Called at the start of each message.  There may be multiple messages
** in a connection.  Any per-message data should be initialized here.
**
** The sender address is also known as the "envelope from" address.
**
** fromargs: Null-terminated SMTP command arguments; fromargs[0] is the
**           sender address.
*/
static sfsistat
sample_envfrom( SMFICTX* ctx, char** fromargs )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_envfrom( \"%s\" )", cd->connection_num, fromargs[0] );

    ++cd->messages;
    cd->recipients = 0;
    cd->headers = 0;
    cd->bodyblocks = 0;
    cd->bodybytes = 0;

    return SMFIS_CONTINUE;
    }


/* sample_envrcpt - handle a RCPT TO:<> command
**
** Called separately for each recipient of a message.
**
** rcptargs: Null-terminated SMTP command arguments; rcptargs[0] is the
**           recipient address.
*/
static sfsistat
sample_envrcpt( SMFICTX* ctx, char** rcptargs )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_envrcpt( \"%s\" )", cd->connection_num, rcptargs[0] );

    ++cd->recipients;

    return SMFIS_CONTINUE;
    }


/* sample_header - handle a header line
**
** Called separately for each header line in a message.
**
** name:  Header field name.
** value: Header vield value, including folded whitespace.  The final CRLF
**        is removed.
*/
static sfsistat
sample_header( SMFICTX* ctx, char* name, char* value )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_header( \"%s\", \"%s\" )", cd->connection_num, name, value );

    ++cd->headers;

    return SMFIS_CONTINUE;
    }


/* sample_eoh - handle the end of the headers
**
** Called once per message after all headers have been processed.
*/
static sfsistat
sample_eoh( SMFICTX* ctx )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_eoh()", cd->connection_num );

    return SMFIS_CONTINUE;
    }


/* sample_body - handle a block of body bytes
**
** The body of the message is filtered via zero or more calls to this
** routine.  Most messages are handled with a single body call, but
** some do take more than one.
**
** bytes: Pointer to the block of body data.
** len:   Length of the block.
*/
static sfsistat
sample_body( SMFICTX* ctx, unsigned char* bytes, size_t len )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_body( %d )", cd->connection_num, (int) len );

    ++cd->bodyblocks;
    cd->bodybytes += len;

    return SMFIS_CONTINUE;
    }


/* sample_eom - handle the end of the message
**
** Called once per message after all body blocks have been processed.
** Any per-message data should be freed both here and in sample_abort(),
** and wherever you return any of SMFIS_REJECT/SMFIS_TEMPFAIL/SMFIS_DISCARD.
*/
static sfsistat
sample_eom( SMFICTX* ctx )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_eom(), %d recipients, %d headers, %d bodyblocks with %d bytes", cd->connection_num, cd->recipients, cd->headers, cd->bodyblocks, (int) cd->bodybytes );

    return SMFIS_CONTINUE;
    }


/* sample_abort - handle the message being aborted
**
** May be called at any time during processing of a message, indicating
** that something went wrong.  Any per-message data should be freed both
** here and in sample_eom(), and wherever you return any of
** SMFIS_REJECT/SMFIS_TEMPFAIL/SMFIS_DISCARD.
*/
static sfsistat
sample_abort( SMFICTX* ctx )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_abort(), %d recipients, %d headers, %d bodyblocks with %d bytes", cd->connection_num, cd->recipients, cd->headers, cd->bodyblocks, (int) cd->bodybytes );

    return SMFIS_CONTINUE;
    }


/* sample_close - handle the connection being closed
**
** Called once at the end of a connection.  Any per-connection data
** should be freed here.
*/
static sfsistat
sample_close( SMFICTX* ctx )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_close(), %d messages", cd->connection_num, cd->messages );

    if ( cd != (struct connection_data*) 0 )
	{
	(void) smfi_setpriv( ctx, (void*) 0 );
	free( (void*) cd );
	}

    return SMFIS_CONTINUE;
    }


/* sample_unknown - unknown / unimplemented command
**
** Called on unknown or unimplemented commands.
*/
static sfsistat
sample_unknown( SMFICTX* ctx, const char* cmd )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_unknown(), '%s'", cd->connection_num, cmd );

    return SMFIS_CONTINUE;
    }


/* sample_data - DATA command filter
**
** Called to allow modification of the message body.
*/
static sfsistat
sample_data( SMFICTX* ctx )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_data()", cd->connection_num );

    return SMFIS_CONTINUE;
    }


/* sample_negotiate - DATA command filter
**
** Called to allow modification of the message body.
*/
static sfsistat
sample_negotiate( SMFICTX* ctx, unsigned long f0, unsigned long f1, unsigned long f2, unsigned long f3, unsigned long* pf0, unsigned long* pf1, unsigned long* pf2, unsigned long* pf3 )
    {
    struct connection_data* cd = (struct connection_data*) smfi_getpriv( ctx );

    syslog( LOG_INFO, "conn# %d - sample_negotiate(), %lxd %lxd %lxd %lxd", cd->connection_num, f0, f1, f2, f3 );

    return SMFIS_ALL_OPTS;
    }
